#!/usr/bin/perl
#
# rndCiv.pl - make randomized FreeCiv rulesets
# $Id: rndCiv.pl,v 1.35 2003/10/17 05:33:58 banjo Exp $
# Copyright (C) 2003, Billy Naylor <banjo@actrix.co.nz>
# 
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#

use strict;

# -----------------------------------------------------------
# Usage

my $helptext = <<EOF;

$0: make randomized FreeCiv rulesets
Usage: $0 [OPTIONS] packagelist

Options:

    -v/-vv/-vvv  verboseness, more 'v's more verbose
    -s/--size    number of techs to have
    -o/--outdir  directory to write to
                 defaults to ./RndCiv/
    -d/--data    data directory
    -t/--techs   must have techs
                 ( use '-' to continue to package list )
    -h/--help    print this help & exit
    -l/--list    list known packages

EOF

# -----------------------------------------------------------
# Configuration

srand (time ^ $$ ^ unpack "%L*", `ps axww | gzip`);

our $debug = 0;

# maxiumum number of techs freeciv allows
my $MAXTECHS = 199;

# How many techs
my $maxtechs = 90;

# data directory
my $datadir = './rulesets';

# output directory
my $outdir = './output';

# -----------------------------------------------------------
# Parse args

my ( @pkgs, $techs, @musthave );

while ( @ARGV ) {
	$_ = shift;

	if ( m/^-+h(elp)?$/ ) {
		print $helptext;
		exit 1;
	}

	elsif ( m/^-+s(ize)?$/ ) {
		die "No size value\n" unless scalar @ARGV;
		my $size = shift;
		die "Unable to have size '$size'\n" unless $size =~ /^\d+$/;
		$maxtechs = $size;
	}

	elsif ( m/^-+(v+)$/ ) {
		$debug = length $1;
		$debug and print "Setting Debug level to $debug\n";
	}

	elsif ( m/^-+t(echs)?$/ ) {
		while ( @ARGV ) {
			$_ = shift;
			last if ( m/^-+$/ );
			push @musthave, $_;
		}
	} 

	elsif ( m/^-+o(outdir)?$/ ) {
		die "No output directory\n" unless scalar @ARGV;
		$outdir = shift;
	}

	elsif ( m/^-+d(ata)?$/ ) {
		die "No data directory given\n" unless scalar @ARGV;
		$_ = shift;
		die "Unable to locate data dir '$_'\n"
		 unless ( -d $_ );
		$datadir = $_;
	}

	elsif ( m/^-+l(ist)?$/ ) {
		die "No data directory found\n" unless ( -d $datadir );
		print join ', ', map{ m/\.(\w+)$/; $1 } qx{ ls $datadir/techs.* };
		print "\n";
		exit 1;
	}

	elsif ( m/^-/ ) {
		next if m/^-+-+$/;
		die "Unknown option '$_'\n";
	}

	else { # packages
		die "Bad package name '$_'\n"
		unless m/^\w+$/;
		push @pkgs, $_;
	}
}

# -----------------------------------------------------------
# Parse tech packages


=pod $techs

  $techs->{
	$techname {
		{flags},
		{prereqs},
	},
  }

=cut

my ( $pkgseen, $pastTop, $pkgcount ) = ( {}, 0, 0 );
push @pkgs, '-';  # top of stack
while ( scalar @pkgs ) {
	my $pkg = shift @pkgs;

	if ( $pkg =~ m/^\-$/ ) {
		$pastTop = 1;
		next;	
	}

	my $filename = "$datadir/techs.$pkg";
	die "Unknown package '$pkg'\n"
	  unless ( -e $filename );
	next if ( defined $pkgseen->{$pkg} );
	$pkgseen->{$pkg} = ++$pkgcount;

	open( TECHS, "< $filename" )
	  or die "Unable to open package $pkg '$filename':$!\n";

	my $linecount = 0;
	for (<TECHS>) {
		$linecount++;
		next if m/^#/;      # ignore comments ( hash in column 1 )
		next if m/^\s*$/;   # and blank lines
		my $errhelp = "$_ [$filename : $linecount]";

		# grab commands
		if ( m/^\+\s+(\w+)\s+(.*?)\s*$/ ) {
			my( $cmd, $arg ) = ( $1, $2 );

			if ( $cmd eq 'read' ) {
				my @choices = &shuffle( split /\s*\|\s*/, $arg );
				( scalar @choices ) or
				 warn "No options listed $errhelp\n" and next;
				push @pkgs, shift @choices;
			}

			elsif ( $cmd eq 'choose' ) {
				next if $pastTop;
				my @choices = &shuffle( split /\s*\|\s*/, $arg );
				( scalar @choices ) or
				 warn "No options listed $errhelp\n" and next;
				push @musthave, shift @choices;
			}

			else {
				warn "Bad command: $errhelp\n";
			}
			next;
		}

		# grab techs
		next unless m/
			^(.*?)			# name
			\:\s+			# name delim
			(\((.*?)\))?	# possible flags
			\s*(.*?)\s*$	# requirements
		/x;
		my ( $name, $flags, @reqs ) = ( $1, $3, split( /\s*,\s*/, $4 ) );
		( scalar @reqs > 1 )
		 or die "Too few prereqs for tech '$name' $errhelp\n";
	
		if ( defined $techs->{$name} ) {
			$techs->{$name}{flags} = $flags
			 if rand() > 0.4;
		}
		else {
			$techs->{$name}{flags} = $flags;
		}
		push @{ $techs->{$name}{prereqs} }, @reqs;
	}
	close TECHS;
}

my @tek = keys %{ $techs };
die "No tech packages loaded\n"
 unless ( scalar @tek );

$maxtechs = ( scalar @tek ) if ( scalar @tek < $maxtechs );
$maxtechs = $MAXTECHS if ( $maxtechs > $MAXTECHS );

# -----------------------------------------------------------
# Build techtree

=pod $techtree

 $techtree->{
	'None' => 1,
	'Name' => $name,
 }

=cut

my $techtree  = { None => { req1 => undef, req2 => undef, node => rndCiv::node->new( 'None' ) } };
my $techcount = 0;

for my $tech ( @musthave ) {
	$debug == 1 and print "Must have tech $tech\n";
	&buildown( $tech )
	 or $debug and print "Unable to build tree down for tech '$tech'\n";
}

for my $tech ( &shuffle( @tek ) ) {
	last if ( $techcount > $maxtechs );
	$debug == 1 and print "Building down from $tech\n";
	&buildown( $tech )
	 or $debug and print "Unable to build tree down for tech '$tech'\n";
}

# -----------------------------------------------------------
# Write our ruleset

unless ( -d $outdir ) {
	mkdir $outdir
	 or die "Unable to make output directory '$outdir':$!\n";
}

open ( FILE, "> $outdir/techs.ruleset" )
 or die "Unable to open '$outdir/techs.ruleset' for writing:$!\n";

print FILE <<EOF;
; created by rndCiv.pl
; total $techcount techs

[datafile]
description="Randomized technology data for Freeciv"
options="1.9"

EOF

my $summary = {
	'None' => {
		pre       => [],
		units     => [],
		buildings => [],
		govs      => [],
	},
};
for my $name ( sort keys %{ $techtree } ) {
	next if $name eq 'None' or $name eq '';
	my ( $req1, $req2, $flags ) = (
		$techtree->{$name}{req1},
		$techtree->{$name}{req2},
		$techs->{$name}{flags}
	);

	my $advance = lc "advance_$name";
	$advance =~ s/ /_/g;

	( $req1, $req2 ) = ( $req2, $req1 )
	 if ( $req1 eq 'None' and $req2 ne 'None' );

	$flags ||= '""';

	print FILE "[$advance]\n"
	 . "name     = _(\"$name\")\n"
	 . "req1     = \"$req1\"\n"
	 . "req2     = \"$req2\"\n"
	 . "flags    = $flags\n\n";

	$summary->{$name}{pre}       = [ $req1, $req2 ];
	$summary->{$name}{units}     = [];
	$summary->{$name}{buildings} = [];
	$summary->{$name}{govs}      = [];
}

close FILE;

# -----------------------------------------------------------
# set up lookup list

my $got = {
	techs       => {},
	governments => {},
	buildings   => {},
	units       => {},
};

for my $k ( keys %{ $techtree } ) {
	$got->{techs}{$k} = 1;
}
undef $techtree;


# -----------------------------------------------------------
# parse government rules

$debug and print "Parsing Government rules\n";
{
	my ( $govs, $options ) = readData( 'governments', keys %{ $pkgseen } );

	my $output;
	for my $g ( keys %{ $govs } ) {
		$govs->{$g}{nogood} == 1 and next;
		$output .= $govs->{$g}{data} . "\n";
		die "No PreReq found for \n'$govs->{$g}{data}'\n"
		 unless $govs->{$g}{data} =~ m/^tech_req\s+=\s+\"(.+?)\"\s*$/m;
		push @{ $summary->{ $1 }{govs} }, $g;
	}

	my $aigoal = '';
	for my $u ( reverse sort( shuffle( @{ $options->{ai_goal} } ) ) ) {
		$u =~ s/^\s*\d+\s+//;
		next unless exists $got->{governments}{$u};
		$aigoal = $u;
		last;
	}
	die "Unable to assign ai_goal\n" unless $aigoal;

	my $ai_tech_hints = '';
	for my $u ( @{ $options->{ai_tech_hints} } ) {
		$u =~ m/^\s*(.+?),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+)\s*$/
		 or warn "Invalid ai_tech_hints '$u'\n" and next;
		next unless exists $got->{techs}{$1};
		$ai_tech_hints .= "    \"$1\", $2, $3, $4, $5\n";	
	}

	# print to file
	open ( FILE, "> $outdir/governments.ruleset" )
	 or die "Unable to open '$outdir/governments.ruleset' for writing:$!\n";

	print FILE <<EOF;
; created by rndCiv.pl

[datafile]
description="Randomized governments data for Freeciv"
options="1.9"

[governments]
default="Despotism"
when_anarchy="Anarchy"
ai_goal="$aigoal"

ai_tech_hints = { "tech", "turns_factor", "const_factor", "get_first", "done"
$ai_tech_hints}

$output

EOF

	close FILE;
}

# -----------------------------------------------------------
# parse buildings rules

$debug and print "Parsing Buildings rules\n";
{
	my ( $buildings, $options ) = readData( 'buildings', keys %{ $pkgseen });

	my $output;
	for my $b ( sort keys %{ $buildings } ) {
		$buildings->{$b}{nogood} == 1 and next;
		$output .= $buildings->{$b}{data} . "\n";
		die "No PreReq found for \n'$buildings->{$b}{data}'\n"
		 unless $buildings->{$b}{data} =~ m/^tech_req\s+=\s+\"(.+?)\"\s*$/m;
		push @{ $summary->{ $1 }{buildings} }, $b;
	}


	# print to file
	open ( FILE, "> $outdir/buildings.ruleset" )
	 or die "Unable to open '$outdir/buildings.ruleset' for writing:$!\n";

	print FILE <<EOF;
; created by rndCiv.pl

[datafile]
description="Randomized buildings data for Freeciv"
options="1.10.1"

$output

; FIXME: remove all of the following when gen-impr implemented...

[b_special]

; Special values:

aqueduct_size=8;
sewer_size=12;

; Techs which modify building effects:
; ### this is bad...

cathedral_plus="Theology"
cathedral_minus="Communism"
colosseum_plus="Electricity"
temple_plus="Mysticism"
EOF

	close FILE;
}

# -----------------------------------------------------------
# parse units rules

$debug and print "Parsing Units rules\n";
{
	my ( $units, $options ) = readData( 'units', keys %{ $pkgseen } );

	my $output;
	for my $u ( sort { $units->{$a}{level} <=> $units->{$b}{level} } keys %{ $units } ) {
		$units->{$u}{nogood} == 1 and next;
		$output .= $units->{$u}{data} . "\n";
		die "No PreReq found for \n'$units->{$u}{data}'\n"
		 unless $units->{$u}{data} =~ m/^tech_req\s+=\s+\"(.+?)\"\s*$/m;
		push @{ $summary->{ $1 }{units} }, $u;
	}

	my ( $partisan_req, $partisan_count ) = ( '', 0 );
	for my $u ( shuffle( @{ $options->{partisan_req} } ) ) {
		next unless exists $got->{techs}{$u};
		$partisan_req .= "\"$u\",";
		$partisan_count++;
		last if $partisan_count == 2;
	}
	$partisan_req =~ s/,$//;

	# print to file
	open ( FILE, "> $outdir/units.ruleset" )
	 or die "Unable to open '$outdir/units.ruleset' for writing:$!\n";

	print FILE <<EOF;
; created by rndCiv.pl

[datafile]
description="Randomized units data for Freeciv"
options="1.9"

[units_adjust]
max_hitpoints=0                 ; 0 means no max
max_firepower=0                 ; 0 means no max
firepower_factor=1              ; convenient for Civ1

$output

[u_specials]
partisan_req = $partisan_req
EOF

	close FILE;
}
 
# -----------------------------------------------------------
# write summary

open ( FILE, "> $outdir/ruleset.summary" )
 or die "Unable to open '$outdir/ruleset.summary' for writing:$!\n";

print FILE "Summary of rndCiv ruleset"
 . "\nDate: " . scalar localtime( time )
 . "\n$techcount techs\n";

for my $t ( sort keys %{ $summary } ) {
	print FILE "\n$t (" . join ', ', @{ $summary->{$t}{pre} };
	print FILE ")";

	scalar @{ $summary->{$t}{units} } and
	print FILE "\n units: ", join ', ', @{ $summary->{$t}{units} };
	scalar @{ $summary->{$t}{buildings} } and
	print FILE "\n buildings: ", join ', ', @{ $summary->{$t}{buildings} };
	scalar @{ $summary->{$t}{govs} } and
	print FILE "\n governments: ", join ', ', @{ $summary->{$t}{govs} };
}

print FILE "\n\n";
close FILE;

exit 1;

# -----------------------------------------------------------
# subs

# The Fisher Yates Shuffle from the Perl Cookbook
sub shuffle {
	my @array = @_;
	for my $i ( 0..$#array ) {
		my $j = int rand( $i + 1 );
		next if $i == $j;
		( $array[$i], $array[$j] ) = ( $array[$j], $array[$i] );
	}
	return @array;
}

# begin to build down to 'None' from a given tech
sub buildown {
	my $tech = shift;
	my ( $tree, %tekList ) = &buildnodesdown( $tech, () )
	 or return 0;
	&readtree( $tree )
	 or return 0;
	return 1;
}

# create a tech tree down to 'None'
# might redefine working techs ??
sub buildnodesdown {
	my ( $name, %tekList ) = @_;

	if ( exists $techtree->{$name} ) {
		$debug > 2 and print "'$name' is in techtree\n";
		return $techtree->{$name}{node};
	}

	unless ( exists $techs->{$name} ) {
		$debug > 1 and print "Unknown tech '$name' called\n";
		return undef;
	}

	# already found 
	if ( exists $tekList{$name} ) {
		$debug > 2 and print "already have $name\n";
		return $tekList{$name}{node};
	}
	$tekList{$name}{node} = undef;

	# choose an available prereq
	my ( $pre1, $pre1name, $hold, $holdname );
	for my $p ( &shuffle( @{ $techs->{$name}{prereqs} } ) ) {
		$debug > 1 and print "$name: Trying prereq $p\n";
		my ( $ret, %newTL ) = &buildnodesdown( $p, %tekList );

		unless ( defined $ret ) {
			$debug > 2 and print "No path found for '$p'\n";
			next;
		}
		next if $ret->contains( $name );

		unless ( $pre1 ) {
			$pre1 = $ret;
			$pre1name = ( ref $pre1 ) ? $pre1->{name} : $pre1;
			next;
		}
		my $pre2 = $ret;
		my $pre2name = ( ref $pre2 ) ? $pre2->{name} : $pre2;

		unless ( $pre1name eq 'None' or $pre2name eq 'None' ) {
			next if ( $pre2name eq $pre1name );
			# i would like to use these constraints
			# but then the tree may not even build!
#			next if ( $pre1->{req1}{name} eq $pre2name );
#			next if ( $pre1->{req2}{name} eq $pre2name );
#			next if ( $pre2->{req1}{name} eq $pre1name );
#			next if ( $pre2->{req2}{name} eq $pre1name );
		}

		if ( $pre2name eq "None" ) {
			#next if ( $holdname ne 'None' );
			( $hold,  $holdname ) = ( $pre2, $pre2name );
			next;
		}
		
		( $hold,  $holdname ) = ( $pre2, $pre2name );
		if ( $pre1->contains( $pre1name )
		  or $pre2->contains( $pre2name ) ) {
			$debug > 2 and print "Holding $pre2name\n";
			next;
		}

		last;
	}

	unless ( $pre1 and $hold ) {
		$debug > 1 and print "Failed to find 2 prereqs for tech '$name'\n";
		return undef
	}

	$debug > 2 and print "New node for $name( $pre1name, $holdname )\n";
	my $node = rndCiv::node->new( $name );
	$node->{req1} = $pre1;
	$node->{req2} = $hold;
	return ( $node, %tekList );
}

# parse sucessful tree into $techtree
sub readtree {
	my $node = shift;
	my $name = ( ref $node ) ? $node->{name} : $node;
	return $name
	  if ( exists $techtree->{$name} );

	$techtree->{$name} = {};
	$techtree->{$name}{node} = $node;
	for my $r ( 'req1', 'req2' ) {
		$techtree->{$name}{$r} = ( ref $node->{$r} ) ?
		 &readtree( $node->{$r} ) : $node->{$r};
	}
	$techcount++;
	return $name;
}

# utility for reading governments, buildings and units data files
sub readData {
	my ( $type, @pkgList ) = @_;
	die "Bad type '$type'\n"
	 unless ( $type =~ m/^(units|governments|buildings)$/ );

	my ( $data, $options ) = ( {}, {} );
	for my $p ( @pkgList ) {
		$debug > 1 and print "Reading package '$datadir/$type.$p'\n";
		unless( open( FILE, "< $datadir/$type.$p" ) ) {
			$debug > 2 and print "units ruleset $datadir/$type.$p not found\n";
			next;
		}

		my ( $buf, $level, $oldlevel ) = ( '', 0, 0 );
		for (<FILE>) {
			next if m/^\#/;
			next if m/^\s*$/;
			next if m/^\s*\;/;
			if ( m/^\[/ and $buf ) {
				$oldlevel ||= $level;
				my $ob = rndCiv::Data->new( $buf, $oldlevel );
				$oldlevel = $level;
				$data->{ $ob->{Name} } = $ob;
				$buf = '';
			}
			elsif ( m/^\+\s+level\s+(\d+)/ ) {
				$oldlevel = $level;
				$level = $1;
				next;
			}
			elsif ( m/^\+\s+(\w+)\s+(.*?)\s*$/ ) {
				push @{ $options->{$1} }, $2;
				next;
			}
			$buf .= $_;
		}
		close FILE;
		if ( $buf ) {
			my $ob = rndCiv::Data->new( $buf, $level );
			$data->{ $ob->{Name} } = $ob;
		}
	}

	return ( &checkdepends( $data, $type ),  $options );
}

# this is called from readdata and checks dependencies are ok
# call in order of gov, building, unit
# loops until no more good things found
sub checkdepends {
	my ( $z, $type ) = @_;
	my $lookup = {
		T => 'techs',
		G => 'governments',
		U => 'units',
		B => 'buildings',
	};

	my $gotone = 1;
	while ( $gotone ) {
		$gotone = 0;
		for my $g ( keys %{ $z } ) {
			next unless $z->{$g}{nogood};
			my $hold = 0;

			while ( $z->{$g}{data} =~ m/\<[TGUB].+?\>/ ) {
				# look for equiv_dupl corner case
				if ( $z->{$g}{data} =~ m/\<[TGUB][^\>]+?,/ ) {
					$z->{$g}{data} =~ s/\<([TGUB])/<X$1/m;
					next;
				}

				$z->{$g}{data} =~ m/\<([TGUB])\s+(.+?)\s*\>/;
				my ( $class, $options ) = ( $1, $2 );
				$z->{$g}{data} =~ s/\<$class/<=$class/; # hide from while loop
				my ( $canbenone, $cancel, $tmp ) = ( 0, 0, '' );

				for my $o ( split /\s*\|\s*/, $options ) {
					$o eq 'None' and $canbenone = 1 and next;  # fallback to "None"
					$o eq 'cancel' and $cancel = 1 and next;   # if fails then delete this line

					$lookup->{$class} eq $type and not exists $z->{$o} and next;
					exists $got->{ $lookup->{$class} }{$o} and $tmp = $o and last;
				}

				if ( $tmp ) {
					$z->{$g}{data} =~ s/\<=[TGUB]\s+.+?\s*\>/"$tmp"/;
				}
				else {
					if ( $canbenone and not $gotone and not $hold ) {
						$z->{$g}{data} =~ s/\<=[TGUB]\s+.+?\s*\>/"None"/;
					}
					elsif ( $cancel ) {
						if ( $class eq 'T' ) {
							$z->{$g}{data} =~ s/^.*?\<=T\s+[^\>]*\bcancel\b.*$//m;
						}
						elsif ( not $gotone ) {
							$z->{$g}{data} =~ s/^.*?\<=$class\s+[^\>]*\bcancel\b.*$//m;
						}
						else {
							$hold = 1;
						}
					}
					else {
						$hold = 1;
						$debug > 1 and print "Holding $class $g\n";
					}
				}
				$z->{$g}{data} =~ s/\<=$class/<==$class/; # hide from while loop & cancel
			}

			unless ( $hold ) {
				$debug and print "Added $g\n";
				$got->{$type}{$g} = 1;
				$z->{$g}{nogood} = 0;
				$gotone = 1;
			}
			$z->{$g}{data} =~ s/\<=+/</g; # unhide
		}
	}

	# now sort out the equiv_dupl corner case
	for my $g ( keys %{ $z } ) {
		while ( $z->{$g}{data} =~ m/\<X([TGUB])\s+(.+?)\s*\>/ ) {
			my ( $class, $options ) = ( $1, $2 );  # $class should be 'B'
			my @keepers;
			for my $o ( split /\s*,\s*/, $options ) {
				next unless exists $got->{ $lookup->{$class} }{$o};
				push @keepers, $o;
			}
			my $buf = ( scalar @keepers ) ? join '", "', @keepers : 'None';
			$z->{$g}{data} =~ s/\<X[TGUB]\s+.+?\s*\>/"$buf"/;
		}
	}

	return $z;
}

# -----------------------------------------------------------
# object class for nodes

package rndCiv::node;

sub new {
	my ( $class, $name ) = @_;
	return bless {
		name => $name,
		req1 => undef,
		req2 => undef,
	}, $class;
}

sub contains {
	my ( $self, $name ) = @_;
	return 0 if $self->{name} eq 'None';
	return 1 if $self->{req1}->contains( $name );	
	return 1 if $self->{req2}->contains( $name );	
	return 0;
}

package rndCiv::Data;

sub new {
	my ( $class, $buf, $level ) = @_;

	my $self = bless {
		data   => $buf,
		Name   => '',
		nogood => 1,
		PreReq => '',
		level  => $level,
	}, $class;

	$buf =~ m/^name\s+=\s+_\(\"(.+?)\"\)\s*$/m
	 or die "No name found for data '$buf'\n";
	$self->{Name} = $1;

	return $self;
}

1;

